#!/usr/bin/perl

# TARDIS
# Tardis Always Recovers Directories In Secret
# Not a time travel contraption but a backup device that appears to get more on 
# to a disc than a disc can hold
# Also replicates (and improves upon though I say so myself) the behaviour of 
# the backup program included with a certain well-known fruit-based computer.

# Fat German Productions 2011

# This program is SpongWare.
# The use of SpongWare is entirely according to the following conditions:
# 1) You may use it or not, I don't care. Just don't winge at me, I hate that
# 2) It is provided 'as is'. Nobody is responsible for anything it might or might not do. 
#    Take your lawyers and stuff 'em up your USB port
# 3) You may distribute it freely, although I would advise against printing it out and 
#    throwing it out of the window, that stuff gets you fined
# 4) You may alter it if you wish, and distribute the altered version if you wish but 
#    you must include this license
# 5) Before installing this software, smile at someone you like
# 6) There is no rule 6
# 7) This notice must be included with any copy of the software you distribute
# 8) Adherence to 'coding standards' is to be regarded with suspicion. 
#    This is Perl, it's not meant to be read anyway ;-)
# 9) Let me reiterate clause 2. If this program goes wrong in any way, for example
#    it deletes all your data or perhaps it wipes the entire internet overnight and 
#    leaves the world a dead place, nobody is responsible. Deal with it.
#    Sometimes, shit happens

# fatgerman@gmail.com

# NOTE requires libfilesys-df-perl      (sudo apt-get install libfilesys-df-perl)
#               libdatetime-perl        (sudo apt-get install libdatetime-perl)
#               libdate-manip-perl      (sudo apt-get install libdate-manip-perl)

use strict;
use warnings;

# Check there isn't already a TARDIS running
my $cmd = "ps -ae | grep TARDIS";
my $result = `$cmd`;
my @lines = split("\\n", $result);
if (scalar @lines > 1) {
    my $line = shift @lines;
    if ($line !~ /defunct/) {
        exit 0;
    }
}


use Getopt::Long;
use File::Basename;

local *PROCESS;
local *FILE;
my $home    = $ENV{HOME};
my $version = "0.06";
my $debug;
getCmdLineOptions();

my $stats_file = "$home/.tardis/stats.log";
my $stats = FatStat->New();
my $fifo = FatFifo->New("/tmp/tardispipe");
my $mod_error = "";
eval { require DateTime } or $mod_error = "ERROR:Perl DateTime module not installed";
eval { require DateTime::Duration } or $mod_error = "ERROR:Perl DateTime module not installed";
eval { require Date::Manip::Date } or $mod_error= "ERROR:Perl Date-Manip module not installed";
eval { require Filesys::Df } or $mod_error = "ERROR:Perl Filesys-Df module not installed";
if ($mod_error ne "") {
    $fifo->Write($mod_error);
    $fifo->Write("FILE:See the config dialog for more info");
    $fifo->Close();
    $stats->Stat("There was an error:", "\nTARDIS could not run because some Perl modules are missing\n".
                                        "\nYou need to install:\n".
                                        "libfilesys-df-perl\t(sudo apt-get install libfilesys-df-perl)\n".
                                        "libdatetime-perl\t(sudo apt-get install libdatetime-perl)\n".
                                        "libdate-manip-perl\t(sudo apt-get install libdate-manip-perl)\n".
                                        "\nSee http://sourceforge.net/p/ktardis/wiki/Home/ for more information.");

    $stats->SaveStats();
    if (-l "$home/.tardis/latest.log") {
        unlink "$home/.tardis/latest.log";
    }       
    exit 0;
}
Filesys::Df->import( 'df' );

# Global variables. Messy but simple
# Global DateTime object which will be the basis for all the date maths later on
# as well as the name of the backup
my $dt = DateTime->from_epoch( 	epoch => time(),
								time_zone => 'local'
							  );
my $datestring = $dt->ymd('-') . 'T' . $dt->hms('_');
my $exclude 			= "$home/.tardis/exclude";
my $options_file 		= "$home/.tardis/tardis.conf";
my $num_logs_to_keep 	= 10;
my $log                = "$home/.tardis/TardisLog_$datestring.log";
my $multipurpose_variable;
my @multipurpose_array;
my %options;
my @backups;
my $hostname;
my $remote_dir;
my $dateformatstring = $dt->day_abbr . " " . $dt->day . " " . $dt->month_abbr . " " . $dt->year . " " . $dt->hms;

# Redirect STDOUT to our log file
if (!defined $debug) {
    tie *STDOUT, 'FatLog', 1, undef, $log, "";
    *STDERR = *STDOUT;
}

autoflush STDOUT 1;

if (!-d "$home/.tardis") {
	system( "mkdir $home/.tardis" );
}

print "TARDIS $version is in operation at $dateformatstring\n";

# Defaults:
# Keep all backups for 24 hours
# After that, keep daily backups for a month
# After that, keep a weekly backup for 12 months
# After that, keep monthly backups
# These options are overriden by the contents of $options_file
$options{source} 						=	"$home/";
$options{dest} 							=	undef;
$options{keep_hourly_backups_for_hours} = 24;
$options{keep_daily_backups_for_months} = 1;
$options{keep_weekly_backups_for_months} = 12;
$options{maximum_used_discspace_percent} = 95;
getfileoptions();

###
### The work starts here
###
$fifo->Write("STATUS:Started");
run_rsync();
skip_hourly_backups();
arrange_daily_backups();
sort_weekly_backups();
sort_monthly_backups();
sort_array_for_deleting();
check_disc_space();
tidy_log_files();
$fifo->Close();
$stats->SaveStats();

###
### And we're done
###

sub run_rsync {
	
	# This actually does the backup. It's so simple even an idiot like me could have come up with it.
	# But I didn't, I nicked the rsync part from somewhere on the web. If it was you please remind me and I'll add a credit

	print "Creating new backup of $options{source} into $options{dest}/back-$datestring\n";
	
	if (!-d $options{source}) {
        print "FATAL! Source directory does not exist\n";
        $fifo->Write("ERROR:Source does not exist");
        $fifo->Close();
        exit(0);
    }
	
	if ($options{dest} !~ m/(.*?)\:(\/.*)/ &&
            !-d $options{dest}) {
        print "FATAL! Destination directory does not exist\n";
        $fifo->Write("ERROR:Destination does not exist");
        $fifo->Close();
        exit(0);
    }

    my $linkdest = "../current";
	my $backupdest = "$options{dest}/back-$datestring";
	
	# Scan stage to work out how many files there are:
    if (open (PROCESS, "rsync -naP --delete-excluded --exclude-from=$exclude --link-dest=$linkdest $options{source}/ $backupdest 2>&1|")) {
        print "Scanning for changed files.....\n";
        $fifo->Write("STATUS:Scanning");
    } else {
        close (PROCESS);
        print "FATAL! It's all gone horribly wrong!\n";
        $fifo->Write("ERROR:Could not start rsync");
        $fifo->Close();
        exit(0);
    }

    my $files = 0;
    while (<PROCESS>) {
        print "DEBUG: $_" if ($debug);
        if (/speedup is/i || /sent \d+ bytes/i || /\/$/ ||
            /sending incremental/i || /created directory/i ||
            /^\s*$/) {
            
        } elsif (/\s+(\d+)\s+\d+\%/) {

        } else {
            $files++;
        }
    }
    close (PROCESS);

    print "    There are $files files to back up\n";
    $fifo->Write("STATUS:Backing up $files files");
    $stats->Stat("Number of Files Backed Up Last Time:", $files);
    
	if (open (PROCESS, "rsync -aP --delete-excluded --exclude-from=$exclude --link-dest=$linkdest $options{source}/ $backupdest.InProgress 2>&1|")) {
		print "rsync is running......\n";
	} else {
		close (PROCESS);
		print "FATAL! It's all gone horribly wrong!\n";
        $fifo->Write("ERROR:Could not start rsync");
        $fifo->Close();
        exit(0);
	}

    my $filessent = 0;
	while (<PROCESS>) {
        print $_;
        if (/speedup is/i || /sent \d+ bytes/i || /\/$/ ||
            /sending incremental/i || /created directory/i ||
            /^\s*$/) {
        } elsif (/\s+(\d+)\s+\d+\%/) {
            # This clearly isn't telling me what I think it's telling me
            # print "Sent $bytessent bytes of $totalbytes\n";
        } else {
            $filessent++;
            chomp;
            chomp;
            $fifo->Write("STATUS:Backing up file $filessent of $files");
            $fifo->Write("FILE:$_");
        }
	}
	close (PROCESS);

    if ($options{dest} =~ m/(.*?)\:(\/.*)/) {
        $hostname = $1;
        $remote_dir = $2;
        print "This is a backup to host $hostname using ssh to directory $remote_dir\n";
        $linkdest = "$remote_dir/current";
        $backupdest = "$remote_dir/back-$datestring";
        system( "ssh $hostname 'mv $backupdest.InProgress $backupdest'" );
        system( "ssh $hostname 'rm -f $linkdest'" );
        system( "ssh $hostname 'ln -s $backupdest $linkdest'" );
        system( "ssh $hostname 'rm -fR $remote_dir/*.InProgress'" );
    } else {
        $linkdest = "$options{dest}/current";
        if (-d $linkdest) {
            system( "rm -f $linkdest" );
        }
        system ("mv $backupdest.InProgress $backupdest");
        system ("ln -s $backupdest $linkdest" );
        system ("rm -fR $options{dest}/*.InProgress");
    }
    $fifo->Write("STATUS:Cleaning Up");
    
    $fifo->Write("IGNORE:Just making sure the buffer is big enough");
    $fifo->Write("IGNORE:Just making sure the buffer is big enough");
    $fifo->Write("IGNORE:Just making sure the buffer is big enough");

}

sub keepbackup {

	# Misleadingly-named function, the purpose of which is both crucial and lost on me. And I wrote it. What hope have you got?
	
	my ($max_months, $max_days, $max_hours,
		 $year, $month, $day, $hour, $minute, $seconds) = @_;
		 		 	
	my $current = DateTime->new(
      year       => $year,
      month      => $month,
      day        => $day,
      hour       => $hour,
      minute     => $minute,
      second     => $seconds
	);

	my $timediff = $dt->subtract_datetime( $current );
	
	print "  is " . $timediff->years() . " years " . $timediff->months() . " months " . $timediff->weeks() . " weeks " . $timediff->days() . " days and " . $timediff->hours() . " hours old\n";

	my $age = DateTime::Duration->new (
		years 		=> 0,
		months		=> $max_months,
		weeks		=> 0,
		days		=> $max_days,
		hours		=> $max_hours,
		minutes		=> 0,
		seconds		=> 0,
	);
	
	my $result = DateTime::Duration->compare($timediff, $age, $dt);
	
	return ($result < 1) ? 1 : 0;
}

sub getweek {

	# Get week. What I do sitting here. Say it out loud.
	
	my ($year, $month, $day, $hour, $minute, $seconds) = @_;

	my $current = DateTime->new(
      year       => $year,
      month      => $month,
      day        => $day,
      hour       => $hour,
      minute     => $minute,
      second     => $seconds
	);
	
	my $wn = $current->week_number();
	
	print "   Week Number is $wn\n";
	
	return $wn;
}

sub getfileoptions {
	
	# Get options. Ahh the 90's. Options were worth something then. I bought an orange.
	if (open (FILE, "<$options_file")) {
		print "Loading options from $options_file\n";
		%options = map { /(.*)\=(.*)\n/; $1 => $2 } <FILE>;
	}
	close (FILE);
	foreach my $k (keys %options) {
		print "Options $k = $options{$k}\n";
	}
	if ($options{source} =~ s/(.*?\:\/\/)//) {
		if ($1 ne "file://") {
            $fifo->Write("ERROR:Protocol $1 is not supported");
            $fifo->Close();
			print "ERROR! Cannot use protocol $1 as a source\n";
            exit(0);
		}
	}
	if ($options{dest} =~ s/(.*?\:\/\/)//) {
		if ($1 ne "file://") {
            $fifo->Write("ERROR:Protocol $1 is not supported");
            $fifo->Close();
			print "ERROR! Cannot use protocol $1 as a destination\n";
            exit(0);
		}
	}
	if (!defined $options{dest}) {
        $fifo->Write("ERROR:Backup Destination is not defined");
        $fifo->Close();
		print "ERROR! Backup destination is not defined!\n";
        exit(0);
	}
}

sub skip_hourly_backups {

	#Scan through our date-sorted list of backups, discarding any less than $options{keep_hourly_backups_for_hours} hours old
	print "Tidying up....\n";
    if (defined $hostname) {
        my $cmd = "ssh $hostname 'ls -l $remote_dir'";
        my $output = `$cmd`;
        my @backup_temp = ();
        foreach my $line (@{[split /\n/, $output]}) {
            if ($line =~ m/\d+\s+(back.*)$/) {
                push @backup_temp, $remote_dir."/".$1;
            }
        }
        @backups = sort {$b cmp $a} @backup_temp;
    } else {
        @backups = sort {$b cmp $a} glob("$options{dest}/*");
    }
    
    if (scalar @backups > 0) {
        $stats->Stat("Oldest Backup:", formatDirToDate($backups[(scalar @backups) - 1]));
    }
    if (scalar @backups > 1) {
        $stats->Stat("Number of Backups:", (scalar @backups) - 2);
    } else {
        $stats->Stat("Number of Backups:", 1);
    }
    
	print "Finding the first backup older than $options{keep_hourly_backups_for_hours} hours\n";
	while (scalar @backups) {	
		my $dir = shift @backups;
		if ($dir =~ m/(\d+)-(\d+)-(\d+)T(\d+)_(\d+)_(\d+)/) {
			print "$dir ";
			my ($year, $month, $day, $hour, $minute, $seconds) = ($1, $2, $3, $4, $5, $6);
			if (keepbackup( 0, 0, $options{keep_hourly_backups_for_hours}, 
							$year, $month, $day, $hour, $minute, $seconds ))
			{
				# Don't do anything
			}
			else
			{
				$multipurpose_variable = $day;
				push @multipurpose_array, $dir;
				last;
			}
		}
	}	
}

sub arrange_daily_backups {
	
	# Now we want to keep one backup per day up to a fixed number of days ago.
	print "Sorting daily backups for up to $options{keep_daily_backups_for_months} months\n";
	while (scalar @backups) {	
		my $dir = shift @backups;
		if ($dir =~ m/(\d+)-(\d+)-(\d+)T(\d+)_(\d+)_(\d+)/) {
			print "$dir\n";
			my ($year, $month, $day, $hour, $minute, $seconds) = ($1, $2, $3, $4, $5, $6);
			if ($day == $multipurpose_variable) {
	#			print " : same day, adding to list\n";
				push @multipurpose_array, $dir;
			}
			else
			{
	#			print " : day has changed.. ";
				sort_array_for_deleting();
				@multipurpose_array = ();
				if (keepbackup( $options{keep_daily_backups_for_months}, 0, 0,
								$year, $month, $day, $hour, $minute, $seconds ))
				{
					$multipurpose_variable = $day;
					push @multipurpose_array, $dir;
				}
				else
				{
	#				print " older than $options{keep_daily_backups_for_months} months, jumping out\n";
					push @multipurpose_array, $dir;
					$multipurpose_variable = getweek($year, $month, $day, $hour, $minute, $seconds);
					last;
				}
			}
		}
	}
}

sub sort_weekly_backups {

	# From now on we want to keep one per week up to the specified number of months
	print "Sorting weekly backups for up to $options{keep_weekly_backups_for_months} months\n";
	while (scalar @backups) {	
		my $dir = shift @backups;
		if ($dir =~ m/(\d+)-(\d+)-(\d+)T(\d+)_(\d+)_(\d+)/) {
			print "$dir\n";
			my ($year, $month, $day, $hour, $minute, $seconds) = ($1, $2, $3, $4, $5, $6);
			my $week = getweek($year, $month, $day, $hour, $minute, $seconds);
			if ($week == $multipurpose_variable) {
	#			print " : same week, adding to list\n";
				push @multipurpose_array, $dir;
			}
			else
			{
	#			print " : week has changed.. ";
				sort_array_for_deleting();
				@multipurpose_array = ();
				if (keepbackup( $options{keep_weekly_backups_for_months}, 0, 0,
								$year, $month, $day, $hour, $minute, $seconds ))
				{
					$multipurpose_variable = $week;
					push @multipurpose_array, $dir;
				}
				else
				{
	#				print " older than $options{keep_weekly_backups_for_months} months, jumping out\n";
					@multipurpose_array = ();
					push @multipurpose_array, $dir;
					$multipurpose_variable = $month;
					last;
				}
			}
		}
	}
}

sub sort_monthly_backups {
	
	# And finally, one per month
	print "Sorting monthly backups forever\n";
	while (scalar @backups) {	
		my $dir = shift @backups;
		if ($dir =~ m/(\d+)-(\d+)-(\d+)T(\d+)_(\d+)_(\d+)/) {
			print "$dir\n";
			my ($year, $month, $day, $hour, $minute, $seconds) = ($1, $2, $3, $4, $5, $6);
			if ($month == $multipurpose_variable) {
	#			print " : same month, adding to list\n";
				push @multipurpose_array, $dir;
			}
			else
			{
	#			print " : month has changed.. ";
				sort_array_for_deleting();
				@multipurpose_array = ();
				$multipurpose_variable = $month;
				push @multipurpose_array, $dir;
			}
		}
	}
}

sub tidy_log_files {
	
	# Tidy up log files
	@backups = sort {$b cmp $a} glob("$home/.tardis/TardisLog_*.log");
	while (scalar @backups > $num_logs_to_keep) {
		my $file = pop @backups;
		system( "rm $file" );
	}
	if (-l "$home/.tardis/latest.log") {
		unlink "$home/.tardis/latest.log";
	}
	if (!defined $debug) {
        system("ln -s $log $home/.tardis/latest.log");
    }
	print "Finished\n\n";
    $fifo->Write("DATE:$dateformatstring");
    $fifo->Write("STATUS:Idle");
    $stats->Stat("Most Recent Backup:", $dateformatstring);
}

sub sort_array_for_deleting {
	# The array contains a list of every backup on the same day, or week, or month - depending on what we're filtering
	# Remove the first element of the array, which will be the most recent and therefore the one we keep
	
	my $keeping = shift @multipurpose_array;
	foreach my $dir (@multipurpose_array) {

		print "      ** Removing $dir **\n";
        $fifo->Write("FILE:Removing ".formatDirToDate($dir));

        if (defined $hostname) {
            system( "ssh $hostname 'rm -fr $dir'" );
        } else {
            system( "rm -fr $dir" );
        }
	}
}

sub check_disc_space {
	
	my $pf = get_used_percent($options{dest});
    my @to_del = ();
	while (defined $pf && $pf > $options{maximum_used_discspace_percent})
	{
        if (defined $hostname) {
            my $cmd = "ssh $hostname 'ls -l $remote_dir'";
            my $output = `$cmd`;
            my @delete_temp = ();
            foreach my $line (@{[split /\n/, $output]}) {
                if ($line =~ m/\d+\s+(back.*)$/) {
                    push @delete_temp, $remote_dir."/".$1;
                }
            }
            @to_del = sort @delete_temp;
        } else {
            @to_del = sort glob("$options{dest}/*");
        }
        my $to_delete = shift @to_del;
		last if !defined $to_delete;
		print "      ** Removing $to_delete **\n";
        $fifo->Write("FILE:Removing ".formatDirToDate($to_delete));
        if (defined $hostname) {
            my $cmd = "ssh $hostname 'rm -fr $to_delete'";
            system($cmd);
        } else {
            system( "rm -fr $to_delete" );
        }
		$pf = get_used_percent($options{dest});
	}

}

sub get_used_percent {
	
    my $pf = undef;
    if (defined $hostname) {
        my $cmd = "ssh $hostname 'df $remote_dir'";
        my $out = `$cmd`;
        if ($out =~ m/(\d+)\%/) {
            $pf = $1;
        }
    } else {
        my $ref = df(shift);
        if(defined($ref)) {
            $pf = $ref->{per};
        }
    }
    if (defined $pf) {
        print "Backup disc is $pf percent full out of a permitted $options{maximum_used_discspace_percent} percent\n";
        $fifo->Write("FILE:Backup disc is $pf% full");
        $stats->Stat("Used Space on Backup Disc:", "$pf%");
        return $pf;
    } else {
        print "ERROR Unable to measure disc space!\n";
        $fifo->Write("ERROR:Unable to measure disc space");
        return undef;
    }
}


sub getCmdLineOptions {

  GetOptions( 'debug' => \$debug);

}

sub formatDirToDate {

    my ($dir) = @_;
    my $datestr = basename($dir);
    $datestr =~ s/back-//;
    $datestr =~ s/T/ /;
    $datestr =~ s/_/:/g;
    my $fmt = new Date::Manip::Date;
    $fmt->parse($datestr);
    return $fmt->printf("%a %e %b %Y %T");

}
  
BEGIN: {
	
	package FatLog;

	local *REDIRECT;

	sub TIEHANDLE {

        my ($class, $copyHandle, $outHandle, $rootname, $path) = @_;

        open (REDIRECT, ">$rootname");

        bless {
            _rname   => $rootname,
            _file    => \*REDIRECT
        }, $class;

	}

	sub READLINE {

        my ($self) = @_;
        my $name = ${$self->{_fname}};
        return $name;

	}

	sub PRINT {

        my ($self, $text) = @_;
        my $logger = $self->{_file};

        $text .= "\n" if ($text !~ /\n$/);
        print $logger $text;

	}

	sub DESTROY {

        my ($self) = @_;
        close ($self->{_file});
        return 1;

	}

	1;

}

BEGIN: {

    package FatFifo;
    use Fcntl;
    use IO::Handle;
    
    local *FIFO;
    
    sub New {
    
        my ($class, $filename) = @_;
        my $print_to_fifo = 1;
        
        if (-p $filename) {
            sysopen(FIFO, $filename, O_NONBLOCK|O_WRONLY) or $print_to_fifo = 0;
            if ($print_to_fifo) {
                print "Using FIFO $filename to communicate with the Plasma widget\n" if ($debug);
                autoflush FIFO 1;
                #FIFO->autoflush(1);
            } else {
                print "Couldn't open FIFO, assuming Plasma widget is not running\n" if ($debug);
            }
        } else {
            print "No FIFO found, assuming Plasma widget is not running\n" if ($debug);
            $print_to_fifo = 0;
            close(FIFO);
        }
        
        bless {
            _flag => $print_to_fifo,
            _handle => \*FIFO
        }, $class;
        
    }
    
    sub Write {
    
        my ($self, $line) = @_;
        if ($self->{_flag}) {
            syswrite $self->{_handle}, $line."!!ENDLINE!!\n";
        }
        if ($debug) {
            print "FIFO : $line\n";
        }
    
    }

    sub Close {

        my ($self) = @_;
        if ($self->{_flag}) {
            close($self->{_handle});
        }
        
    }
    
    1;
    
}

BEGIN: {

    package FatStat;
    
    local *STATS;
    
    sub New {
    
        my ($class) = @_;
        
        bless {
            _stats => {}
        }, $class;
        
    }
    
    sub Stat {
    
        my ($self, $name, $value) = @_;
        $self->{_stats}->{$name} = $value;
    
    }

    sub SaveStats {
    
        my ($self) = @_;
        print "======================================\n";
        print "STATS:\n";
        open (STATS, ">$stats_file");
        foreach my $key (keys $self->{_stats}) {
            print "$key ".$self->{_stats}->{$key}."\n";
            print STATS "$key ".$self->{_stats}->{$key}."\n";
        }
        close (STATS);
        print "======================================\n";
    
    }
    
    1;
    
}