# -*- coding: utf-8 -*-

# TARDIS configuration and monitoring widget for Plasma

# Fat German Productions 2011

# This program is SpongWare.
# The use of SpongWare is entirely according to the following conditions:
# 1) You may use it or not, I don't care. Just don't winge at me, I hate that
# 2) It is provided 'as is'. Nobody is responsible for anything it might or might not do. 
#    Take your lawyers and stuff 'em up your USB port
# 3) You may distribute it freely, although I would advise against printing it out and 
#    throwing it out of the window, that stuff gets you fined
# 4) You may alter it if you wish, and distribute the altered version if you wish but 
#    you must include this license
# 5) Before installing this software, smile at someone you like
# 6) There is no rule 6
# 7) This notice must be included with any copy of the software you distribute
# 8) Adherence to 'coding standards' is to be regarded with suspicion. 
# 9) Let me reiterate clause 2. If this program goes wrong in any way, for example
#    it deletes all your data or perhaps it wipes the entire internet overnight and 
#    leaves the world a dead place, nobody is responsible. Deal with it.
#    Sometimes, shit happens

# fatgerman@gmail.com

# I wrote this to teach myself python. I wish I hadn't

from PyQt4.QtCore import *
from PyQt4.QtGui import *
from PyQt4 import uic
from PyKDE4 import plasmascript
from PyKDE4.plasma import *
from PyKDE4.kio import *
from PyKDE4.kdeui import *
from PyKDE4.kdecore import *
import os
from crontab import CronTab
from tardis import *
import subprocess
import time
import re

DEBUG = 0

class TardisWidget(plasmascript.Applet):
    VERSION = "0.07"
    
    def __init__(self, parent, args=None):
        plasmascript.Applet.__init__(self, parent)
        #self.setApplet(Plasma.Applet(parent, []))
        self.parent = parent

    def init(self):
        self.homedir = os.getenv("HOME") + "/"
        self.configdir = self.homedir + ".tardis"
        self.configfile = self.configdir + "/tardis.conf"
        self.triggerfile = self.configdir + "/trigger.log"
        self.excludefile = self.configdir + "/exclude"
        self.statsfile = self.configdir + "/stats.log"
        debugfile = self.configdir + "/widget.log"
        self.baricon = "tardis-widget"
        self.minspast = 0
        self.runevery = 1
        self.aniTimer = QTimer()
        self.fifoTimer = QTimer()
        self.debug = None
        self.fifo = None
        self.remainingdata = ""
        self.canRun = True
        self.icontheme = "Dark"
        self.icons = {}
        self.icons['Dark'] = dict([('flash', 'tardis-light'), ('normal', 'tardis-widget')])
        self.icons['Light'] = dict([('flash', 'tardis-white-light'), ('normal', 'tardis-widget-white')])
        self.isError = False
        
        self.doTheMessyStuff()
        
        self.setAspectRatioMode(Plasma.IgnoreAspectRatio)
        self.theme = Plasma.Svg(self)
        self.theme.setImagePath("widgets/background")
        self.setBackgroundHints(Plasma.Applet.DefaultBackground)
        
        self.widget=TardisThing(self)
        self.widget.init()
        self.setGraphicsWidget(self.widget)    
        self.setPassivePopup(True)
        
        if (self.canRun):
            self.isRunning = False
            self.getIconTheme();
            self.isEnabled = self.checkCronTab()
            self.widget.updateCheckBox(self.isEnabled)
            self.connect(self.widget.runbutton, SIGNAL("clicked()"), self.runNowButtonClicked)
            self.connect(self.aniTimer, SIGNAL("timeout()"), self.flashPopupIcon)
            self.connect(self.fifoTimer, SIGNAL("timeout()"), self.readFromFifo)
            self.connect(self.applet, SIGNAL("appletDestroyed(Plasma::Applet*)"), self.doExit)
            self.fifoTimer.start(2000);            
            self.initTooltip()
            self.checkLastDate()
        else:
            self.applet.setPopupIcon(KIcon("dialog-error"))

    def initTooltip(self):
        if (self.formFactor() != Plasma.Planar): 
            self.tooltip = Plasma.ToolTipContent() 
            self.tooltip.setImage(KIcon(self.baricon))
            self.tooltip.setMainText( "TARDIS Backup")
            Plasma.ToolTipManager.self().setContent(self.applet, self.tooltip)
            #Plasma.ToolTipManager.self().registerWidget(self.applet)

    #@pyqtSlot(name='toolTipAboutToShow')
        #def toolTipAboutToShow(self):
            #pass

    def createConfigurationInterface(self, parent):        
        self.config_widget = QWidget(parent)
        self.config_ui = uic.loadUi(str(self.package().filePath('ui', 'ihatekde.ui')), self.config_widget)
        self.config_ui.backup_folder.setMode( KFile.Modes( KFile.Directory ) )
        self.config_ui.backup_destination.setMode( KFile.Modes( KFile.Directory ) )
        parent.addPage(self.config_widget, "TARDIS", "tardis-widget" )
        self.connect(self.config_ui.frankSpencer, SIGNAL("leftClickedUrl(const QString&)"), self.openWikiPage)

        self.config_ui.backup_folder.setPath( self.homedir )
        self.isEnabled = self.checkCronTab()
        self.config_ui.tardisEnabled.setCheckState(Qt.Checked if self.isEnabled else Qt.Unchecked)

        self.retention_widget = QWidget(parent)
        self.retention_ui = uic.loadUi(str(self.package().filePath('ui', 'retention.ui')), self.retention_widget)
        parent.addPage(self.retention_widget, "Retention", "user-trash" )

        self.exclude_widget = QWidget(parent)
        self.exclude_ui = uic.loadUi(str(self.package().filePath('ui', 'exclude.ui')), self.exclude_widget)
        self.exclude_ui.excludelist.setButtons( KEditListWidget.Buttons( KEditListWidget.Add | KEditListWidget.Remove ))
        parent.addPage(self.exclude_widget, "Exclusions", "inode-directory-128" )

        self.stats_widget = QWidget(parent)
        self.stats_ui = uic.loadUi(str(self.package().filePath('ui', 'stats.ui')), self.stats_widget)
        self.stats_ui.statsbox.setText(getFile(self.statsfile))
        self.stats_ui.logButton.setIcon(KIcon("text-x-log"))
        if (self.isError):
            parent.addPage(self.stats_widget, "Stats", "dialog-error" )
        else:
            parent.addPage(self.stats_widget, "Stats", "dialog-information" )
        if (os.path.isfile(self.configdir+"/latest.log")):
            self.connect(self.stats_ui.logButton, SIGNAL("clicked()"), self.logButtonClicked)
        else:
            self.stats_ui.logButton.setEnabled(False)

        self.loadConfig()
        if (self.icontheme == "Dark"):
            self.config_ui.radioDark.setChecked(Qt.Checked)
        else:
            self.config_ui.radioLight.setChecked(Qt.Checked)
        self.connect(parent, SIGNAL("okClicked()"), self.updateConfig)
        self.connect(self.config_widget, SIGNAL('destroyed(QObject*)'), self.configWidgetDestroyed)
        return self.config_widget

    def openWikiPage(self, url):
        try:
            out = subprocess.check_call( ["kde-open", url] )
        except (OSError, CalledProcessError):
            self.debugprint("Couldn't Open WebPage. Weasels.")

    def logButtonClicked(self):
        try:
            out = subprocess.check_call( ["kde-open", "file://"+self.configdir+"/latest.log"] )
        except (OSError, CalledProcessError):
            self.debugprint("Couldn't Open Log File. Weasels.")

    def updateConfig(self):
        self.icontheme = "Dark" if (self.config_ui.radioDark.isChecked()) else "Light"
        try:
            fp = open(self.configfile,"w")
            fp.write( "source=" + self.config_ui.backup_folder.url().url() + "\n")
            fp.write( "dest=" + self.config_ui.backup_destination.url().url() + "\n" )
            fp.write( "keep_hourly_backups_for_hours=" + str(self.retention_ui.keephourly.value()) + "\n")
            fp.write( "keep_daily_backups_for_months=" + str(self.retention_ui.keepweekly.value()) + "\n")
            fp.write( "keep_weekly_backups_for_months=" + str(self.retention_ui.keepmonthly.value()) + "\n")
            fp.write( "maximum_used_discspace_percent=" + str(self.retention_ui.percentfull.value()) + "\n")
            fp.write( "runevery=" + str(self.config_ui.runevery.value()) + "\n")
            fp.write( "minspast=" + str(self.config_ui.minspast.value()) + "\n")
            fp.write( "icontheme=" + self.icontheme + "\n")
            fp.close()
        except IOError:
            self.debugprint( "Error opening config file for write")

        try:
            fp = open(self.excludefile, "w")
            stringlist = self.exclude_ui.excludelist.items()
            for string in stringlist:
                if (string != ""):
                    fp.write( "- /" + string + "\n" )
            fp.close()
        except IOError:
            self.debugprint("Error opening exclude file for write")

        self.isEnabled = self.checkCronTab()
        if (self.config_ui.tardisEnabled.checkState() == Qt.Checked):
            if (self.isEnabled == False):
                self.enableCronTab()
                self.widget.updateCheckBox(self.isEnabled)        
        else:
            if (self.isEnabled == True):
                self.disableCronTab()
                self.widget.updateCheckBox(self.isEnabled)

        if (self.config_ui.runevery.value() != self.runevery | self.config_ui.minspast.value() != self.minspast):
            self.runevery = self.config_ui.runevery.value()
            self.minspast = self.config_ui.minspast.value()
            if (self.isEnabled):
                self.disableCronTab()
                self.enableCronTab()
        
        self.flashPopupIcon()

    def configWidgetDestroyed(self):
        self.exclude_widget = None
        self.config_widget = None
        self.retention_widget = None
        self.stats_widget = None
        self.exclude_ui = None
        self.config_ui = None
        self.retention_ui = None
        self.stats_ui = None
        
    def getIconTheme(self):
        for line in getFile(self.configfile).split('\n'):
            t = line.partition("=")
            if t[0] == "icontheme":
                self.icontheme = t[2].rstrip()
        self.flashPopupIcon()
        
    def loadConfig(self):
        for line in getFile(self.configfile).split('\n'):
            t = line.partition("=")
            if t[0] == "source":
                url = KUrl( t[2].rstrip() );
                self.config_ui.backup_folder.setUrl( url )
            elif t[0] == "dest":
                url = KUrl( t[2].rstrip() );
                self.config_ui.backup_destination.setUrl( url )
            elif t[0] == "keep_hourly_backups_for_hours":
                self.retention_ui.keephourly.setValue( int(t[2]) )
            elif t[0] == "keep_daily_backups_for_months":
                self.retention_ui.keepweekly.setValue( int(t[2]) )
            elif t[0] == "keep_weekly_backups_for_months":
                self.retention_ui.keepmonthly.setValue( int(t[2]) )
            elif t[0] == "maximum_used_discspace_percent":
                self.retention_ui.percentfull.setValue( int(t[2]) )
            elif t[0] == "runevery":
                self.config_ui.runevery.setValue( int(t[2]) )
                self.runevery = int(t[2])
            elif t[0] == "minspast":
                self.config_ui.minspast.setValue( int(t[2]) )
                self.minspast = int(t[2])
            elif t[0] == "icontheme":
                self.icontheme = t[2].rstrip()
            else:
                pass

        for line in getFile(self.excludefile).split('\n'):
            if not line.isspace():
                self.exclude_ui.excludelist.insertItem( line.rstrip().lstrip("- /"), -1 )

    def checkCronTab(self):
        tab = CronTab()
        num = tab.find_command('TARDIS.pl')
        if (len(num) > 0):
            return True
        else:
            return False

    def enableCronTab(self):
        tab = CronTab()
        cmd = tab.new(command = unicode(self.package().path()) + "contents/code/TARDIS.pl",comment="TARDIS incremental backup")
        cmd.minute().on(self.minspast)
        cmd.hour().every(self.runevery)
        tab.write()
        self.isEnabled = self.checkCronTab()

    def disableCronTab(self):
        tab = CronTab()
        tab.remove_all('TARDIS.pl')
        tab.write()
        self.isEnabled = self.checkCronTab()

    def runNowButtonClicked(self):
        if not (self.isRunning):
            self.process = subprocess.Popen(unicode(self.package().path()) + "contents/code/TARDIS.pl")

    def checkLastDate(self):
        for line in getFile(self.statsfile).split('\n'):
            self.widget.updateLastDate(line.rstrip())
    
    def readFromFifo(self):
        self.fifoTimer.stop()
        if (self.fifo !=None):
            try:
                self.data = ""
                self.data = self.remainingdata+os.read(self.fifo, 1024)
                self.remainingdata = ""
                for line in self.data.split('\n'):
                    if line != "" and line != None:
                        if (re.search('!!ENDLINE!!', line) == None):
                            self.remainingdata = self.remainingdata + line
                        else:
                            r = line.replace('!!ENDLINE!!', '').rstrip().partition(':')
                            if r[0] == "STATUS":
                                if r[2] == "Started":
                                    self.isRunning = True
                                    self.isError = False
                                    self.widget.updateLabel(r[2])
                                    self.widget.updateInfo("")
                                    if (self.aniTimer.isActive() == False):
                                        self.aniTimer.start(2000)
                                elif r[2] == "Idle":
                                    self.isRunning = False
                                    if self.isError == False:
                                        self.widget.updateLabel(r[2])
                                    # Closing the fifo allows the other end to quit
                                    os.close(self.fifo)
                                    self.fifo = os.open("/tmp/tardispipe", os.O_RDONLY | os.O_NONBLOCK)
                                else:
                                    self.widget.updateLabel(r[2])
                            elif r[0] == "FILE":
                                self.widget.updateInfo(r[2])
                            elif r[0] == "DATE":
                                self.widget.updateLastDate("Most Recent Backup:"+r[2])
                            elif r[0] == "ERROR":
                                self.isRunning = False
                                self.isError = True
                                self.widget.updateLabel("ERROR : "+r[2])
                                self.aniTimer.stop()
                                self.applet.setPopupIcon(KIcon("dialog-error"))
                            else:
                                pass
                    
            except (IOError, OSError):
                # A read error from the FIFO is normal if TARDIS has it open but
                # there is nothing to be read. Hence we just ignore it.
                self.debugprint("FIFO Error: Got "+self.data)
                pass
            if (self.isRunning):
                self.fifoTimer.start(250)
            else:
                self.fifoTimer.start(1000)

    def flashPopupIcon(self):
        if (self.isRunning == True):
            if (self.baricon == self.icons[self.icontheme]['normal']):
                self.baricon = self.icons[self.icontheme]['flash']
            else:
                self.baricon = self.icons[self.icontheme]['normal']
        else:
            self.baricon = self.icons[self.icontheme]['normal']
            self.aniTimer.stop()
        self.setPopupIcon(self.baricon)
        self.widget.runbutton.setIcon(self.baricon)

    def doExit(self):
        if (self.aniTimer.isActive() == True):
            self.aniTimer.stop()
        if (self.fifoTimer.isActive() == True):
            self.fifoTimer.stop()        
        if (DEBUG == 2):
            self.debug.close()
        if (self.fifo != None and self.canRun):
            os.close(self.fifo)
            try:
                os.unlink("/tmp/tardispipe")
            except:
                pass
        self.widget.deleteLater()
        
    def debugprint(self,out):
        if (DEBUG == 2):
            try:
                self.debug.write(out + "\n")
            except IOError:
                print "File write error with : " + out
        elif (DEBUG == 1):
            print out
        else:
            pass
        
    def doTheMessyStuff(self):
        if not os.path.exists(self.configdir):
            try:
                os.makedirs(self.configdir)
            except (OSError, IOError):
                print "Failed to create config directory. This is probably bad"

        if (DEBUG == 2):
            try:
                self.debug = open(debugfile, "w")
            except IOError:
                print "Couldn't even create debug file. This is appalling"

        try:
            out = subprocess.check_call( ["chmod", "+x", unicode(self.package().path()) + "contents/code/TARDIS.pl"] )
        except:
            self.debugprint("Error with chmod")

        try:
            out = subprocess.check_call( ["xdg-icon-resource", "install", "--size", "128", unicode(self.package().path()) + "contents/icons/tardis-128.png", "tardis-widget"])
        except (OSError, CalledProcessError):
            self.debugprint("Error Installing Icon")

        try:
            out = subprocess.check_call( ["xdg-icon-resource", "install", "--size", "128", unicode(self.package().path()) + "contents/icons/tardis-flash-128.png", "tardis-light"])
        except (OSError, CalledProcessError):
            self.debugprint("Error Installing Icon")

        try:
            out = subprocess.check_call( ["xdg-icon-resource", "install", "--size", "128", unicode(self.package().path()) + "contents/icons/tardis-white-128.png", "tardis-widget-white"])
        except (OSError, CalledProcessError):
            self.debugprint("Error Installing Icon")

        try:
            out = subprocess.check_call( ["xdg-icon-resource", "install", "--size", "128", unicode(self.package().path()) + "contents/icons/tardis-white-flash-128.png", "tardis-white-light"])
        except (OSError, CalledProcessError):
            self.debugprint("Error Installing Icon")

        if (os.path.exists("/tmp/tardispipe")):
            self.setHasConfigurationInterface(False)
            self.canRun = False;
        else:
            self.setHasConfigurationInterface(True)
            try:
                os.mkfifo("/tmp/tardispipe")
                self.fifo = os.open("/tmp/tardispipe", os.O_RDONLY | os.O_NONBLOCK)
            except:
                self.debugprint("Error opening FIFO")
                self.fifo = None

def CreateApplet(parent):
    return TardisWidget(parent)

def getFile(file):
    s = ""
    try:
        with open(file) as f:
            s = f.read()
            f.close()
    except:
        pass
    return s
